var $ = require('bling'),
	Os = require('os'),
	Fs = require('fs'),
	Handlebars = require('handlebars'),
	Shell = require('shelljs'),
	Child = require('./child'),
	Worker = require('./worker'),
	Http = require('./http'),
	Opts = require('./opts'),
	log = $.logger("[herd]"),
	Herd = function Herd(opts) {
		var self = $.extend(this, Herd.defaults(opts)),
			webUpdate = null, pool = null, i = 0, j = 0;
		log("Spawning", self.exec.count, "children")
		self.children = new Array(self.exec.count);
		for( ; i < self.exec.count; i++) {
			self.children[i] = new Child(self, i)
		}

		if( self.workers.length > 0 ) {
			i = 0
			self.workers = $(self.workers).map(function(w) {
				log("Spawning", w.count, "workers")
				var a = [];
				for( j = 0; j < w.count; j++) {
					a.unshift( new Worker(w, i++) )
					a[0].spawn()
				}
				return a
			}).flatten()
		}

		self.shepherdId = [Os.hostname(), self.admin.port].join(":")

		// connect to rabbitmq
		if( self.rabbitmq.enabled && self.rabbitmq.url != null && self.rabbitmq.url != '') {
			require('./amqp').connect(self.rabbitmq.url).wait(function(err, context) {
				if( err ) {
					log("Failed to connect to rabbitmq at:", self.rabbitmq.url, "error:")
					log(err.stack)
					self.rabbitmq.publish = self.rabbitmq.subscribe = $.identity;
				} else {
					self.rabbitmq.context = context;
					self.rabbitmq.publish = function(message) {
						var pub = self.rabbitmq.context.socket("PUB")
						pub.connect(self.rabbitmq.exchange, function() {
							pub.end(JSON.stringify(message), 'utf8')
						})
					}
					self.rabbitmq.subscribe = function(handler) {
						var sub = self.rabbitmq.context.socket("SUB")
						sub.connect(self.rabbitmq.exchange, function() {
							var on_data = function(data) {
								handler(JSON.parse(String(data)))
							}
							sub.on('data', on_data)
							sub.on('drain', on_drain)
						})
					}
					// broadcast that we are online
					self.rabbitmq.publish({ op: "status", status: "online", from: self.shepherdId })
				}
			})
		} else {
			self.rabbitmq.publish = self.rabbitmq.subscribe = $.identity;
		}

		// on SIGINT or SIGTERM, kill everything and die
		["SIGINT", "SIGTERM"].forEach(function(sig) {
			process.on(sig, function() {
				self.killAll("SIGKILL").then(function() {
					process.exit(0);
				}, function(err) {
					console.log(err);
					process.exit(1);
				})
			})
		});

		// on SIGHUP, just reload all child procs
		process.on("SIGHUP", function() {
			self.rollingRestart()
			self.workerRestart()
		});

		// hook up a broadcaster to announce our demise
		["exit", "SIGINT", "SIGTERM"].forEach(function(sig) {
			process.on(sig, function() {
				log("Got signal:", sig)
				self.rabbitmq.publish({ op: "status", status: "offline", from: self.shepherdId, signal: sig })
			})
		});

		function webRoute(method, url, handler) { $.publish("http-route", method, url, handler) }
		function rabbitRoute(pattern, handler) { $.publish("amqp-route", self.rabbitmq.exchange, pattern, handler) }

		function plain(f) {
			return function(req, res) {
				res.contentType = "text/plain";
				res.send = function(status, content, enc) {
					enc = enc || 'utf8';
					res.statusCode = status;
					res.end(content, enc);
				}
				return f(req, res);
			}
		}

		// Worker status via REST
		webRoute("get", "/workers", plain(function(req, res) {
			res.send(200, $.toString(self.getWorkerStatus()))
		}))
		// Worker restart via REST
		webRoute("get", "/workers/restart", plain(function(req, res) {
			self.workerRestart()
			res.redirect(302, "/workers")
		}))

		// Child status via REST endpoint:
		webRoute("get", "/children", plain(function(req, res) {
			self.getChildrenStatus().then(function(list) {
				res.send(200, JSON.stringify(list))
			}, function(err) {
				res.send(500, JSON.stringify(err.stack));
			})
		}))

		// Restart all via REST endpoint:
		webRoute("get", "/children/restart/all", plain(function(req, res) {
			res.send(200, "Restarting all...")
			self.rollingRestart()
		}))
		webRoute("get", "/children/restart", plain(function(req, res) {
			res.send(200, "Restarting all...")
			self.rollingRestart()
		}))

		// Restart a single child process via REST endpoint:
		webRoute("get", "/children/restart/:index", plain(function(req, res) {
			var index = parseInt(req.params.index, 10),
				child = self.children[index],
				fail = function(err) {
					if( err && 'stack' in err ) res.send(500, err.stack)
					else res.send(500, JSON.stringify(err))
				}
				finish = function(text) {
					res.send(200, text)
				}

			if( child == null ) return fail("No such child.")

			child.kill().wait(self.restart.gracePeriod, function(err) {
				if( err != null ) return fail(err)
				else finish("Child "+index+" killed, now it should auto-restart.")
			})
		}))

		// Update all via REST endpoint
		var webUpdate = plain(function(req, res) {
			self.update().then(function(output) {
				res.send(200, output)
			}, function(err) {
				res.send(500, JSON.stringify(err));
			})
		})
		webRoute("get", "/children/update", webUpdate)
		webRoute("post", "/children/update", webUpdate) // listen for POST too, to make webhook support trivial

		// publish an "update" message to everyone over rabbitMQ
		var webBroadcast = plain(function(req, res) {
			try {
				self.rabbitmq.publish({ op: "update" }) // tell everyone to update (including ourself)
				res.send(200, "Broadcast sent.")
			} catch( err ) {
				res.send(500, err.stack)
			}
		})
		webRoute("get", "/children/broadcast-update", webBroadcast)
		webRoute("post", "/children/broadcast-update", webBroadcast)

		// Restart all via AMQP message:
		rabbitRoute({ to: self.shepherdId, op: "restart", arg: "all" }, function (msg) {
			self.rollingRestart()
		})

		// Restart one via AMQP message:
		rabbitRoute({ to: self.shepherdId, op: "restart" }, function (msg) {
			try {
				var index = parseInt(msg.arg, 10)
			} catch(err) {
				return log("Invalid arg in rabbitmq 'restart' message:", msg.arg)
			}
			var child = self.children[index]
			if( child == null ) log("No such child:", index)
			else child.kill()
		})

		// Update via AMQP message:
		rabbitRoute({ op: "update" }, function (msg) {
			self.update()
		})

		// Response to a status query via AMQP message:
		rabbitRoute({ op: "ping" }, function (msg) {
			self.getChildrenStatus().then(function(status) {
				status = { children: status, workers: self.getWorkerStatus() }
				var response = { op: "pong", ts: $.now, dt: $.now - msg.ts, from: self.shepherdId, status: status }
				log("Sending PONG message...", response)
				self.rabbitmq.publish(response)
			})
		})

	} // end constructor Herd

module.exports = Herd

// make sure a herd object has all the default stuff
Herd.defaults = function(opts) {

	opts = $.extend(Object.create(null), opts)
	// the above has two effects: allows calling without arguments
	// and ensures that the opts object can't be polluted by Object.prototype

	opts.exec = $.extend(Object.create(null), {
		cd: ".",
		cmd: "node index.js",
		count: Math.max(1, Os.cpus().length - 1),
		port: 8000, // a starting port, each child after the first will increment this
		portVariable: "PORT", // and set it in the env using this variable
		env: {}
	}, opts.exec)
	opts.exec.port = parseInt(opts.exec.port, 10)
	opts.exec.count = parseInt(opts.exec.count, 10)

	opts.restart = $.extend(Object.create(null), {
		maxAttempts: 5, // failing five times fast is fatal
		maxInterval: 10000, // in what interval is "fast"?
		gracePeriod: 3000, // how long to wait for a forcibly killed process to die
		timeout: 10000, // how long to wait for a newly launched process to start listening on it's port
	}, opts.restart)

	opts.git = $.extend(Object.create(null), {
		remote: "origin",
		branch: "master",
		command: "git pull {{remote}} {{branch}} || git merge --abort"
	}, opts.git)

	opts.git.command = Handlebars.compile(opts.git.command)
	opts.git.command.inspect = function(level) {
		return '"' + opts.git.command({ remote: "{{remote}}", branch: "{{branch}}" }) + '"'
	}

	opts.rabbitmq = $.extend(Object.create(null), {
		enabled: true,
		url: "amqp://localhost:5672",
		exchange: "shepherd"
	}, opts.rabbitmq)

	switch(Os.platform()) {
		case 'darwin':
			opts.nginx = $.extend(Object.create(null), {
				config: "/usr/local/etc/nginx/conf.d/shepherd_pool.conf",
				reload: "launchctl stop homebrew.mxcl.nginx && launchctl start homebrew.mxcl.nginx" // only used if nginx.config is set to a writeable filename
			}, opts.nginx)
			break;
		case 'linux':
			opts.nginx = $.extend(Object.create(null), {
				config: "/etc/nginx/conf.d/shepherd_pool.conf",
				reload: "/etc/init.d/nginx reload" // only used if nginx.config is set to a writeable filename
			}, opts.nginx)
			break;
		default:
			opts.nginx = $.extend(Object.create(null), {
				config: null,
				reload: "echo"
			}, opts.nginx)
			log("Unknown platform:", Os.platform(), "nginx configuration defaults will be meaningless.")
	}

	opts.admin = $.extend(Object.create(null), { // the http server listens for REST calls and webhooks
		port: 9000
	}, opts.admin)

	// "worker" can be set in the config json
	// but it just morphs into the "workers" array
	if( $.is('object', opts.worker) ) {
		opts.workers = [ opts.worker ]
		delete opts.worker
	}
	// and then "worker" becomes read-only (see: "workers")
	$.defineProperty(opts, "worker", {
		get: function() {
			return opts.workers[0]
		}
	})

	if( ! $.is('array', opts.workers) ) {
		opts.workers = []
	}

	opts.workers = $(opts.workers).map(function(w) {
		return $.extend(Object.create(null), {
			count: 1,
			cd: ".",
			cmd: "echo No cmd specified"
		}, w);
	})

	if( Opts.verbose ) log("Using configuration:", require('util').inspect(opts))

	return opts;
}

Herd.prototype.listen = function() {
	Http.listen(this.admin.port)
}

Herd.prototype.setStatus = function(status) {
	var self = this;
	self.rabbitmq.publish({ op: "status", status: status, from: self.shepherdId })
}

Herd.prototype.update = function() {
	var self = this,
		child = $(self.children).coalesce(),
		p = $.Promise(),
		cmd = 'bash -c "cd ' + self.exec.cd + ' && ' + self.git.command(self.git) + '"'

	if( child == null ) return p.reject("No children to update")

	$.Promise.exec(cmd).wait(function(err, output) {
		if( err != null ) return p.reject(err)
		p.resolve(output)
		self.rollingRestart()
	})

	return p
}

Herd.prototype.workerRestart = function() {
	var self = this;
	$(self.workers).each(function(worker) {
		if( worker.process != null ) {
			log("Killing old worker:", worker.process.pid)
			$.Promise.allChildrenOf(worker.process.pid).wait(function(err, pids) {
				log("Killing all worker children:", pids)
				Shell.exec("kill -15 " + $(pids).join(" "))
				// will cause the worker to restart itself (as from un-natural death)
			})
		} else {
			log("Launching new worker...")
			worker.spawn()
		}
	})
}

Herd.prototype.rollingRestart = function(from, done) {
	var self = this,
		from = from || 0,
		next = from + 1,
		done = done || $.Promise(),
		fail = function(msg) {
			done.reject(msg)
		}
	if( Opts.verbose ) log("Rolling restart:", from)
	try {
		if( from < 0 ) {
			fail('invalid index')
		} else if( from >= self.children.length ) {
			done.resolve()
		} else if( self.children[from] == null ) {
			fail('invalid child index: '+from)
		} else if( self.children[from].process == null ) {
			self.children[from].spawn().wait(function(err) {
				if( err != null ) return fail(err)
				self.rollingRestart(next, done)
			})
		} else {
			self.children[from].kill().wait(self.restart.gracePeriod, function(err) {
				if( err != null ) return fail(err)
				self.children[from].started.wait(function(err) {
					if( err != null ) return fail(err)
					self.rollingRestart(next, done)
				})
			})
		}
	} catch (_err) {
		fail(_err)
	}
	return done;
}

var signals = {
	SIGKILL: 9,
	SIGTERM: 15,
	SIGINT: 2,
	SIGHUP: 1,
}

Herd.prototype.killAll = function(signal) {
	var self = this, p = $.Progress(self.children.length + self.workers.length);
	if( ! $.is("number", signal) ) {
		if( ! (signal in signals) )
			return;
		else
			signal = signals[signal];
	}
	$(self.children).each(function(child) {
		if( child && child.process ) {
			$.Promise.allChildrenOf(child.process.pid).then(function(pids) {
				log("Killing children:", pids)
				Shell.exec("kill -" + signal + " " + $(pids).join(" "))
				p.finish(1)
			}, function(err) { p.reject(err) })
			child.process = null
		}
	})
	$(self.workers).each(function(worker) {
		if( worker && worker.process ) {
			$.Promise.allChildrenOf(worker.process.pid).then(function(pids) {
				log("Killing workers:", pids)
				Shell.exec("kill -" + signal + " " + $(pids).join(" "))
				p.finish(1)
			}, function(err) { p.reject(err) })
			worker.process = null
		}
	})
	return p
}

Herd.prototype.writeConfig = function(fail_timeout) {
	var pool, self = this,
		fail_timeout = fail_timeout || "10s";
	if( self.nginx.config != null ) {
		log("Writing nginx configuration to file:", self.nginx.config)
		pool = "upstream shepherd_pool {\n"
			+ self.children.map(function(child) {
				return "\tserver 127.0.0.1:"+child.port+" max_fails=1 fail_timeout="+fail_timeout+";\n"
			}).join('') + "}"
		Fs.writeFile(self.nginx.config, pool, function(err) {
			if( err != null ) log("Failed to write nginx config file:", err)
			else {
				log(self.nginx.reload)
				$.Promise.exec("bash -c '" + self.nginx.reload +"'").wait(function(err) {
					if( err != null ) log("Failed to reload nginx:", err)
				})
			}
		})
	}
}

Herd.prototype.getWorkerStatus = function() {
	return $(this.workers).map(function(w) {
		if( w.process && w.process.pid ) {
			return w.process.pid
		} else {
			return "DEAD"
		}
	}).toArray()
}

Herd.prototype.getChildrenStatus = function() {
	var self = this, p = $.Promise();
	log("Getting status...")
	$.Promise.collect($(self.children).select('getResources').call()).wait(function(err, list) {
		if( err != null ) return p.reject(err)
		var i = 0, pid, port, cpu;
		for( ; i < list.length; i++) {
			port = self.children[i].port
			if( self.children[i].process != null ) {
				pid = self.children[i].serverPid
			} else {
				pid = "DEAD"
			}
			cpu = list[i].split(' ')
				.map(function(x) { return parseFloat(x) })
			list[i] = [pid, port, cpu[0], cpu[1]]
		}
		log("Status:", list)
		p.resolve(list)
	})
	return p
}

